/**
License : CC SA-BY
Arduino UNO motorized robotic arm by Ted Lien v1.0 2016.3.2
0,1 ： RX／TX
2,3,4：右輪（2,4控制方向，3控制速度）
6,7,8：右輪（7,8控制方向，6控制速度）
5：真空馬達
9：電磁閥
10：LED
10,11：超聲波感測器
12,13：藍牙RX／TX
14（A0）：未用到
15,16,17(A1,A2,A3)：左爪右伺服馬達
18,19(A4,A5)：I2C，未用到
*/

#include <Wire.h>
#include <EEPROM.h>
#include <Servo.h> 
#include <SoftwareSerial.h>

const int SERVOS = 3;
const int WHEEL_LEFT_PIN1 = 8;
const int WHEEL_LEFT_PIN2 = 7;
const int WHEEL_RIGHT_PIN1 = 4;
const int WHEEL_RIGHT_PIN2 = 2;
const int LEFT_SPEED_PIN = 6;  //pwm to control left wheel speed
const int RIGHT_SPEED_PIN = 3;  //pwm to control right wheel speed
const int VACUUM = 9; 
const int SWITCH = 5; 
const int LEDPIN = 11;
const int TRIG = 10;  // ultrasound sensor
const int ECHO = 11;  // ultrasound sensor
const int LEFT = 1, RIGHT = 2;
const int OPEN = 1, CLOSE = 0;
const int RANGE = 60;  // maximum cm for autograb

Servo myservo[SERVOS];
SoftwareSerial mySerial(12, 13);     // hc06 bluetooth rx/tx

int PIN[SERVOS] = {15,16,17};  // A1, A2, A3 as D15, D16, D17 (right/clipper/left)
int value[SERVOS];
int INITANGLE[SERVOS] = {90,90,90};
int MAX[SERVOS] = {135,125,135};
int MIN[SERVOS] = {20,80,45};
int servo_moving[SERVOS] = {0,0,0};
int from_angle[3] = {90,90,90}, to_angle[3]={40,130,140}, high_angle[3] = {120, 80, 90}; 
int gripper_angle[2] = {90, 135};  // open/close angle
boolean vacuumOn = false;
boolean isGrabed = false;
const int changeAngle = 70;  // the bigger, the more angle it rotates a time for autograb
                             // normally 50~100 would be good
void setup() {
  //while(1);
  Serial.begin(9600);
  mySerial.begin(38400);
  init_Pins();  // set all pins to OUTPUT except D10, A0, A4, A5
  Wire.begin();

  resetServos();            // initialize Servos
  //servoTest();              // +/- 15% test 
  //while(1);
  //lightupLed();             // test Led at pin 11
  
  //handleMotor(1);
  digitalWrite(LEFT_SPEED_PIN, HIGH);
  digitalWrite(RIGHT_SPEED_PIN, HIGH);
  
  //autoGrab();
  //while(1);
}

void loop() {
  int cmd = checkmySerial();
  if (cmd) Serial.println(cmd);
  if (cmd) go(cmd);
  moveArm();
}

//set motor speed , range 0~100 , 100 at full
void setMotorSpeed(int left, int right){
  analogWrite(LEFT_SPEED_PIN, int(left*255/100));
  analogWrite(RIGHT_SPEED_PIN, int(right*255/100));
}

// set all pins to OUTPUT except SoftSerial(12, 13), unused A0/D14 and I2C(18, 19)
void init_Pins(){  
  for (int i = 2; i <= 10; i++) pinMode(i, OUTPUT); 
  for (int i = 15; i <= 17; i++) pinMode(i, OUTPUT); 
  pinMode(11, INPUT);  
  digitalWrite(10, HIGH);
}

void lightupLed(){
  digitalWrite(LEDPIN, HIGH);
}

void resetServos(){
  for (int i = 0; i < SERVOS; i++){
    myservo[i].attach(PIN[i]);
    myservo[i].write(INITANGLE[i]);
    value[i] = 0;
  }  
}
// test 3 servos within reachable angles
void servoTest(){
  int range = 15;
  for (int i = 0; i < 3; i++){
    for(int j = INITANGLE[i]; j <INITANGLE[i]+range; j++){
      myservo[i].write(j);
      delay(15);
    } 
    for(int j = INITANGLE[i]+range; j > INITANGLE[i]-range; j--){
      myservo[i].write(j);
      delay(15);
    } 
    for(int j = INITANGLE[i]-range; j < INITANGLE[i]; j++){
      myservo[i].write(j);
      delay(15);
    } 
  }
}

void handleMotor(int para){
  //para: 0/1/2/3/4/5/6 : stop wheel/forward/backward/left/right/left_bt/right_bt
  //      10/11/12  : stop compressor/comp-absorb/comp-spit
  switch(para){
    case 0:  //stop both wheels
      //Serial.println("stop both wheels");
      digitalWrite(WHEEL_LEFT_PIN1, LOW);
      digitalWrite(WHEEL_LEFT_PIN2, LOW);
      digitalWrite(WHEEL_RIGHT_PIN1, LOW);
      digitalWrite(WHEEL_RIGHT_PIN2, LOW);
      analogWrite(LEFT_SPEED_PIN, 255); // half speed when turn
      analogWrite(RIGHT_SPEED_PIN, 255);
      break;
    case 1:  //forward
      //Serial.println("forward");
      digitalWrite(WHEEL_LEFT_PIN1, LOW);
      digitalWrite(WHEEL_LEFT_PIN2, HIGH);
      digitalWrite(WHEEL_RIGHT_PIN1, LOW);
      digitalWrite(WHEEL_RIGHT_PIN2, HIGH);
      break;
    case 2:  //backward
      //Serial.println("backward");
      digitalWrite(WHEEL_LEFT_PIN1, HIGH);
      digitalWrite(WHEEL_LEFT_PIN2, LOW);
      digitalWrite(WHEEL_RIGHT_PIN1, HIGH);
      digitalWrite(WHEEL_RIGHT_PIN2, LOW);
      break;
    case 3:  //left turn
      turn(LEFT);       
      break;
    case 4:  //right turn     
      turn(RIGHT); 
      break;  
    case 5:  //left turn
      turn_and_wait_stop(LEFT);    // for body control   
      break;
    case 6:  //right turn     
      turn_and_wait_stop(RIGHT);   // for body control    
      break;        
    default: //stop all
      digitalWrite(WHEEL_LEFT_PIN1, LOW);
      digitalWrite(WHEEL_LEFT_PIN2, LOW);
      digitalWrite(WHEEL_RIGHT_PIN1, LOW);
      digitalWrite(WHEEL_RIGHT_PIN2, LOW);
      break;
  }  
}

// dir 1/2 : left/right turn a bit at a time until stop
void turn_and_wait_stop(int dir){
  while(!mySerial.available()){
    handleMotor(0);
    delay(changeAngle*2);
    if (dir == LEFT){
      digitalWrite(WHEEL_LEFT_PIN1, HIGH);//LOW
      digitalWrite(WHEEL_LEFT_PIN2, LOW);
      digitalWrite(WHEEL_RIGHT_PIN1, LOW);
      digitalWrite(WHEEL_RIGHT_PIN2, HIGH);  
    } else {
      digitalWrite(WHEEL_LEFT_PIN1, LOW);
      digitalWrite(WHEEL_LEFT_PIN2, HIGH);
      digitalWrite(WHEEL_RIGHT_PIN1, HIGH);//LOW
      digitalWrite(WHEEL_RIGHT_PIN2, LOW);  
    }
    delay(changeAngle);
  }
}

// dir 1/2 : left/right turn
void turn(int dir){
  analogWrite(LEFT_SPEED_PIN, 200); // half speed when turn
  analogWrite(RIGHT_SPEED_PIN, 200);
  if (dir == LEFT){
    digitalWrite(WHEEL_LEFT_PIN1, HIGH);//LOW
    digitalWrite(WHEEL_LEFT_PIN2, LOW);
    digitalWrite(WHEEL_RIGHT_PIN1, LOW);
    digitalWrite(WHEEL_RIGHT_PIN2, HIGH);  
  } else {
    digitalWrite(WHEEL_LEFT_PIN1, LOW);
    digitalWrite(WHEEL_LEFT_PIN2, HIGH);
    digitalWrite(WHEEL_RIGHT_PIN1, HIGH);//LOW
    digitalWrite(WHEEL_RIGHT_PIN2, LOW);  
  }
  
}

// speed factor of both wheel, 0~100, 100 as full speed
void readEEPROM(){
  Serial.print(" 0 : ");
  Serial.println(EEPROM.read(0));
  Serial.print(",,, 1 : ");
  Serial.println(EEPROM.read(1));
}

// use ultrasound to get cm distance, using (11,14) as (TRIGger, ECHO)
int getDistance(){
  float duration, distance;
  
  digitalWrite(TRIG, HIGH);
  delayMicroseconds(1000);
  digitalWrite(TRIG, LOW);
  duration = pulseIn (ECHO, HIGH);
  distance = (duration/2)/29;
  return distance;
}

// read command from mySerial and return 2-digits as cmd/para
int checkmySerial() {
  String str = "";
  
  if (mySerial.available()){
    for (int i = 0 ; i < 2; i++){
      char tmp = (char)mySerial.read();
      //Serial.print(tmp);
      if (tmp < '0' || tmp > '9') return 0; 
      Serial.print(tmp);
      Serial.print(" , ");
      str += tmp;
      //str += (char)mySerial.read(); 
    }
    return str.toInt();
  } else {
    return 0;
  }
}

// auto Vacuum on or off
void toggleVacuum(){
  if (vacuumOn) {
    vacuumOn = false;
    digitalWrite(SWITCH, LOW);
    digitalWrite(VACUUM, LOW);    
  } else {
    vacuumOn = true;
    digitalWrite(SWITCH, HIGH);
    delay(200);
    digitalWrite(VACUUM, HIGH);  
  }
}

// manual Vacuum on or off
void vacuum(int i){
  if (i) {
    digitalWrite(SWITCH, HIGH);
    delay(200);
    digitalWrite(VACUUM, HIGH);      
  } else {
    digitalWrite(SWITCH, LOW);
    digitalWrite(VACUUM, LOW);   
  }
}

/*
command : 2-digits , cmd/para
cmd : 
1 : car ,  11/12/13/14 forward/backward/left/right
2 : servo 20,21,22,23,24,25,26 stop/left+,-/clipper+,-/right+,-
3 : auto grab things
4 : auto toggle vacuum on/off
5 : manual toggle vacuum , 51/50 as on/off
6 : 61/60 LED on/off
7 : read distance
9 : stop all
*/

void go(int value){
  int cmd = value / 10;
  int para = value % 10;
  Serial.print(value);
  Serial.print(" , ");
  Serial.print(cmd);
  Serial.print(" / ");
  Serial.println(para);   
 
  if (cmd == 1 ){
    handleMotor(para);       // para: 0/1/2/3/4 : stop/forward/backward/left/right
  } else if (cmd ==2){
    switch (para) {          // servo 0/2 : right / gripper servo
      case 0 :
        servo_moving[0] = 0;  // stop both servos
        servo_moving[1] = 0;
        servo_moving[2] = 0;
        break;
      case 1 : 
        servo_moving[0] = -1;
        break;
      case 2 :
        servo_moving[0] = 1;
        break;
      case 3 :
        if (isGrabed) {
          myservo[1].write(gripper_angle[OPEN]);  // open it
          isGrabed = false;
        } else {
          myservo[1].write(gripper_angle[CLOSE]);  // close it
          isGrabed = true;
        }
        break;
      case 5 :
        servo_moving[2] = -1;
        break;
      case 6 :
        servo_moving[2] = 1;
        break;  
    }    
  } else if (cmd == 3) {      // auto grab/drop
    if (!isGrabed) {
      resetServos();  //reset servo to {90,90,90}
      delay(1000);
    }
    autoGrab();     
  } else if (cmd == 4) {       // auto vacuum on/off 
    toggleVacuum();    
  } else if (cmd == 5) {       // manual vacuum on/off for Scratch2/mBlock
    if (para) vacuum(1);
    else vacuum(0);            
  } else if (cmd == 6) {       // LED on/off
    if (para) digitalWrite(LEDPIN, HIGH);
    else digitalWrite(LEDPIN, LOW);  
  } else if (cmd == 7){
    // read distance      
  } else if (cmd == 9) {              // stop everything
    stopAll();
  }         
}

// auto grab based on ultrasound sensor
void autoGrab(){
  int d;
  boolean hasObject = false;  
  int signal_lost = 0;
  
  if(!isGrabed) {   
    while(1) {   
      handleMotor(0);
      delay(changeAngle*3);
      d = getDistance();
      Serial.println(d);
      if (d < 15){
        hasObject = true;
        signal_lost = 0;
        handleMotor(2);
        delay(changeAngle/2);
      } else if (d >= 15 && d <= 18){ 
        signal_lost = 0;        
        handleMotor(4);
        delay(changeAngle);  // before stop , right turn a little
        handleMotor(0);  
        break;
      } else if (d > 18 && d < RANGE){
        signal_lost = 0;
        hasObject = true;
        handleMotor(1);
        delay(changeAngle/2);
      } else if (hasObject){  // detected but signal lost, left turn a bit if no signal twice
        signal_lost++;
        if (signal_lost > 4){
          signal_lost = 0;
          handleMotor(3);
          delay(changeAngle*2);
          hasObject = false;
        }
      } else {     
        signal_lost = 0;
        handleMotor(4);
        delay(changeAngle/2);
      }
      if(mySerial.available()) {  // stop motor before halt the autograb
        handleMotor(0);  
        return; 
      }
    }
    grab();
    isGrabed = true;
  } else {
    myservo[1].write(gripper_angle[OPEN]);
    isGrabed = false;
    delay(200);
  }
}

// grab thing
void grab(){
  myservo[1].write(gripper_angle[OPEN]);
  delay(200);
  armfromto(from_angle, to_angle);
  delay(500);
  myservo[1].write(gripper_angle[CLOSE]);
  delay(1000);
  armfromto(to_angle, high_angle);
  delay(200);
}

// move by blue tooth for left/right arm
void moveArm(){
  int angle_org, cur_angle[3];
  
  for (int i = 0; i < SERVOS; i++){   
    cur_angle[i] = myservo[i].read();
  }
  
  for (int i = 0; i < SERVOS; i++){    
    if (i==1) continue; // skip gripper , only process left/right arm
    if (servo_moving[i] != 0){  
      angle_org = cur_angle[i];
      cur_angle[i] += servo_moving[i];
      //angle_new = angle_new > MAX[i] ? --angle_new : angle_new;
      //angle_new = angle_new < MIN[i] ? ++angle_new : angle_new;
      if (goodAngle(cur_angle)){  // rotate servo only when angle changed
        myservo[i].write(cur_angle[i]);
        //servo_moving[i] = 0;
        delay(15);
        Serial.print(i);
        Serial.print(":");
        Serial.print(cur_angle[i]);
        Serial.print(" ");
        
      } else {
        cur_angle[i] = angle_org;
      }
      Serial.println();
    }
  }
}

void stopAll(){
  handleMotor(0);                       //stop motor
  for (int i = 0; i < SERVOS; i++){     // stop servos
    myservo[i].write(INITANGLE[i]);
  }
  digitalWrite(VACUUM, LOW);   // stop compressor
  digitalWrite(SWITCH, LOW); 
}

void stopAllServos(){
  for (int i = 0; i < SERVOS; i++){
    servo_moving[i] = 0;
  }
}

// check if angle of left/right out of servo range
boolean goodAngle(int* ang){
  int a[3][4] = {
                  {90,135,60,120},
                  {60,89,45,150},
                  {45,59,90,150}
                };
  
  //if (ang[1] > 125 || ang[1] < 80) return false;  // clipper 
  for (int i = 0; i < 3; i++){  // left/right servo
    if (ang[0] >= a[i][0] && ang[0] <= a[i][1]){                //left servo
      if (ang[2] >= a[i][2] && ang[2] <= a[i][3]) return true;  // right servo
    } 
  }
  return false;
}

// rotate all servos at a time smoothly by using "for" loop and delay(30) 
// only consider left/right servo, ie 0 and 2 
void armfromto(int *arrf, int *arrt){
  int maxAngles = 0;
  float lp[3];
      
  maxAngles = max(abs(arrt[0] - arrf[0]), abs(arrt[2] - arrf[2]));
  maxAngles = maxAngles < 1 ? 1 : maxAngles;  

  for (int i = 0; i < 3; i++){    
    if (i == 1) continue;
    lp[i] = float(arrt[i] - arrf[i])/float(maxAngles);
  }
  
  for (int j = 1; j <= maxAngles; j++){
    for (int i = 0; i < 3; i++){
      if (i == 1) continue;
      myservo[i].write(arrf[i]+j*lp[i]);      
    }
    delay(20);
  } 
}


