/*************************************************************************
* File Name          : uno_mblock_firmware.ino
* Author             : Ted Lien
* Version            : V1.0
* Date               : 03/16/2016
* Description        : Arduino UNO Firmware for Makeblock mBlock  
* License            : CC-BY-SA 3.0
* Origins            : http://www.makeblock.cc/
**************************************************************************/
#include <Wire.h>
#include <SoftwareSerial.h>
//#include <LiquidCrystal_I2C.h>
#include <Servo.h>

Servo servos[8];  
//#define LCD_ROWS 4
//#define LCD_COLUMN 20
//LiquidCrystal_I2C lcd(0x27, LCD_COLUMN, LCD_ROWS);
SoftwareSerial mySerial(12,13);


typedef struct MeModule
{
    int device;
    int port;
    int slot;
    int pin;
    int index;
    float values[3];
} MeModule;

union{
    byte byteVal[4];
    float floatVal;
    long longVal;
}val;

union{
  byte byteVal[8];
  double doubleVal;
}valDouble;

union{
  byte byteVal[2];
  short shortVal;
}valShort;
MeModule modules[12];
int analogs[8]={A0,A1,A2,A3,A4,A5,A6,A7};

String mVersion = "0a.01.102";
boolean isAvailable = false;
boolean isBluetooth = false;

const int LEN = 100;  // max buffer to store command from mBlock
char buffer[LEN];
char bufferBt[LEN];
byte index = 0;
byte dataLen;
byte modulesLen=0;
boolean isStart = false;
unsigned char irRead;
char serialRead;

#define DIGITAL 30
#define ANALOG 31
#define PWM 32
#define SERVO_PIN 33
#define TONE 34
#define PULSEIN 37
#define ULTRASONIC_ARDUINO 36
#define LCD 41
#define TIMER 50

#define GET 1
#define RUN 2
#define RESET 4
#define START 5

float angleServo = 90.0;
int servo_pins[8]={0,0,0,0,0,0,0,0};
unsigned char prevc=0;
double lastTime = 0.0;
double currentTime = 0.0;
uint8_t keyPressed = 0;
int lineNumber = 0, LINEWIDTH = 20;


void setup(){  
  Serial.begin(38400);
  mySerial.begin(38400);
  resetPins();
  flash5();  
  //lcdBegin();
  //delay(100);
}
void loop(){
  readSerial();
  if(isAvailable){
    unsigned char c = serialRead&0xff;
    if(c==0x55&&isStart==false){
     if(prevc==0xff){
      index=1;
      isStart = true;
     }
    }else{
      prevc = c;
      if(isStart){
        if(index==2){
         dataLen = c; 
        }else if(index>2){
          dataLen--;
        }
        writeBuffer(index,c);
      }
    }
     index++;
     if(index>LEN){
      index=0; 
      isStart=false;
     }
     if(isStart&&dataLen==0&&index>3){         
        isStart = false;
        parseData(); 
        index=0;  
     } 
  }
}

// reset pins to default
void resetPins(){
  for(int i = 2; i <= 11; i++){
    pinMode(i, OUTPUT);
    digitalWrite(i, LOW);
  }
}

//flash led 5 times
void flash5(){
  //flash pin 10 LED 5 times
  for (int i = 0; i < 5; i++){
    digitalWrite(10, HIGH);
    delay(100);
    digitalWrite(10, LOW);
    delay(100);
  } 
}

// initial lcd
/*
void lcdBegin(){
  lcd.begin();
  lcd.backlight();
  lcd.print("Hello mBlock!");
}
*/

unsigned char readBuffer(int index){
 return isBluetooth?bufferBt[index]:buffer[index]; 
}
void writeBuffer(int index,unsigned char c){
 if(isBluetooth){
  bufferBt[index]=c;
 }else{
  buffer[index]=c;
 } 
}
void writeHead(){
  writeSerial(0xff);
  writeSerial(0x55);
}
void writeEnd(){
  Serial.println(); 
}
void writeSerial(unsigned char c){
 Serial.write(c);
 mySerial.write(c);
}
void readSerial(){
  isAvailable = false;
  
  if(Serial.available()>0){
    isAvailable = true;
    isBluetooth = false;
    serialRead = Serial.read();
  }
  if(mySerial.available()>0){
    isAvailable = true;
    isBluetooth = true;
    serialRead = mySerial.read();
  }
}
/*
ff 55 len idx action device port  slot  data a
0  1  2   3   4      5      6     7     8
*/
void parseData(){
  isStart = false;
  int header = readBuffer(2);
  int idx = readBuffer(3);
  int action = readBuffer(4);
  int device = readBuffer(5);


  if (header == 2) {
    resetPins();
    return;
  }
  
  switch(action){
    case GET:{
        writeHead();
        writeSerial(idx);
        readSensor(device);
        writeEnd();
     }
     break;
     case RUN:{
       runModule(device);
       callOK();
     }
      break;
      case RESET:{
        callOK();
      }
     break;
     case START:{
        //start
        callOK();
      }
     break;
  }
}
void callOK(){
    writeSerial(0xff);
    writeSerial(0x55);
    writeEnd();
}
void sendByte(char c){
  writeSerial(1);
  writeSerial(c);
}
void sendString(String s){
  int l = s.length();
  writeSerial(4);
  writeSerial(l);
  for(int i=0;i<l;i++){
    writeSerial(s.charAt(i));
  }
}
void sendFloat(float value){ 
     writeSerial(0x2);
     val.floatVal = value;
     writeSerial(val.byteVal[0]);
     writeSerial(val.byteVal[1]);
     writeSerial(val.byteVal[2]);
     writeSerial(val.byteVal[3]);
}
void sendShort(double value){
     writeSerial(3);
     valShort.shortVal = value;
     writeSerial(valShort.byteVal[0]);
     writeSerial(valShort.byteVal[1]);
}
void sendDouble(double value){
     writeSerial(2);
     valDouble.doubleVal = value;
     writeSerial(valDouble.byteVal[0]);
     writeSerial(valDouble.byteVal[1]);
     writeSerial(valDouble.byteVal[2]);
     writeSerial(valDouble.byteVal[3]);
}
short readShort(int idx){
  valShort.byteVal[0] = readBuffer(idx);
  valShort.byteVal[1] = readBuffer(idx+1);
  return valShort.shortVal; 
}
float readFloat(int idx){
  val.byteVal[0] = readBuffer(idx);
  val.byteVal[1] = readBuffer(idx+1);
  val.byteVal[2] = readBuffer(idx+2);
  val.byteVal[3] = readBuffer(idx+3);
  return val.floatVal;
}
void runModule(int device){
  //0xff 0x55 0x6 0x0 0x1 0xa 0x9 0x0 0x0 0xa
  int port = readBuffer(6);
  int pin = port;
  switch(device){
   case DIGITAL:{
     pinMode(pin,OUTPUT);
     int v = readBuffer(7);
     digitalWrite(pin,v);
   }
   break;
   case PWM:{
     pinMode(pin,OUTPUT);
     int v = readBuffer(7);
     analogWrite(pin,v);
   }
   break;
   case TONE:{
     pinMode(pin,OUTPUT);
     int vl = readBuffer(7);
     int vh = readBuffer(8);
     int dl = readBuffer(9);
     int dh = readBuffer(10);
     tone(pin,vl+vh*256,dl+dh*256);
   }
   break;
   case SERVO_PIN:{
     int v = readBuffer(7);
     if(v>=0&&v<=180){
       Servo sv = servos[searchServoPin(pin)];
       sv.attach(pin);
       sv.write(v);
     }
   }
   break;
   case TIMER:{
    lastTime = millis()/1000.0; 
   }
   break;
   /*
   case LCD:{
     int len = readBuffer(6);
     int lines = len/LCD_COLUMN;     
     lcd.clear();
     for(int line = 0; line <= lines; line++){
       for(int i = 0; i < LCD_COLUMN; i++){
         int idx = line*LCD_COLUMN+i;
         if (idx < len){
           lcd.setCursor(i, line);
           lcd.print(char(readBuffer(idx+7)));
         } else return;
       }
     }
   }
   break;   
   */
  }
}

int searchServoPin(int pin){
    for(int i=0;i<8;i++){
      if(servo_pins[i] == pin){
        return i;
      }
      if(servo_pins[i]==0){
        servo_pins[i] = pin;
        return i;
      }
    }
    return 0;
}
void readSensor(int device){
  /**************************************************
      ff 55 len idx action device port slot data a
      0  1  2   3   4      5      6    7    8
  ***************************************************/
  float value=0.0;
  int port,slot,pin;
  port = readBuffer(6);
  pin = port;
  switch(device){
   case  DIGITAL:{
     pinMode(pin,INPUT);
     sendFloat(digitalRead(pin));
   }
   break;
   case  ANALOG:{
     pin = analogs[pin];
     pinMode(pin,INPUT);
     sendFloat(analogRead(pin));
   }
   break;
   case  PULSEIN:{
     int pw = readShort(7);
     pinMode(pin, INPUT);
     sendShort(pulseIn(pin,HIGH,pw));
   }
   break;
   case ULTRASONIC_ARDUINO:{
     int trig = readBuffer(6);
     int echo = readBuffer(7);
     pinMode(trig,OUTPUT);
     digitalWrite(trig,LOW);
     delayMicroseconds(2);
     digitalWrite(trig,HIGH);
     delayMicroseconds(10);
     digitalWrite(trig,LOW);
     pinMode(echo, INPUT);
     sendFloat(pulseIn(echo,HIGH,30000)/58.0);
   }
   break;
   case TIMER:{
     sendFloat((float)currentTime);
   }
   break;
  }
}
